/*
 * braze.c - bootstrap methods
 *
 * Copyright (C) 2014 Henry So, Jr.
 *
 * This file is part of Braze.
 *
 * Braze is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Braze is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Braze.  If not, see <http://www.gnu.org/licenses/>.
 */

#define _GNU_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include "braze.h"

static char **filenames;
static gboolean display_version = FALSE;
gboolean listen = FALSE;
gboolean show_system = FALSE;
gboolean show_patches = FALSE;
gboolean show_connections = TRUE;

#ifndef NDEBUG
gboolean debug_memory = FALSE;
#endif

static GOptionEntry options[] = {
	{ "version", 'v', 0, G_OPTION_ARG_NONE, &display_version, "Show version information and exit", NULL },
	{ "listen", 'l', 0, G_OPTION_ARG_NONE, &listen, "Listen for port changes", NULL },
	{ "show-patches", 'p', 0, G_OPTION_ARG_NONE, &show_patches, "Show desired patch connections", NULL },
	{ "hide-changes", 'c', G_OPTION_FLAG_REVERSE, G_OPTION_ARG_NONE, &show_connections, "Hide patch connection changes", NULL },
	{ "show-system", 's', 0, G_OPTION_ARG_NONE, &show_system, "Show system messages", NULL },
#ifndef NDEBUG
	{ "debug-memory", 0, 0, G_OPTION_ARG_NONE, &debug_memory, "Show allocation summary after each reconfiguration", NULL },
#endif
	{ G_OPTION_REMAINING, 0, 0, G_OPTION_ARG_FILENAME_ARRAY, &filenames, NULL, "FILE" },
	{ NULL },
};

void
abend(int exit_code, const char *__restrict __format, ...) {
	va_list argp;
	va_start(argp, __format);
	vfprintf(stderr, __format, argp);
	va_end(argp);

	exit(exit_code);
}

static void
print_version() {
	printf( "Braze - version 1.0\n"
			"\n"
			"Copyright (C) 2014 Henry So, Jr.\n"
			"\n"
			"This program is free software: you can redistribute it and/or modify\n"
			"it under the terms of the GNU General Public License as published by\n"
			"the Free Software Foundation, either version 3 of the License, or\n"
			"(at your option) any later version.\n"
			"\n"
			"This program is distributed in the hope that it will be useful,\n"
			"but WITHOUT ANY WARRANTY; without even the implied warranty of\n"
			"MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n"
			"GNU General Public License for more details.\n"
			"\n"
			"You should have received a copy of the GNU General Public License\n"
			"along with this program.  If not, see <http://www.gnu.org/licenses/>.\n" );
}

int
main(int argc, char **argv) {
#ifndef NDEBUG
	g_mem_set_vtable(glib_mem_profiler_table);
#endif
	// parse options
	GError *error = NULL;
	GOptionContext *context = g_option_context_new("- scripted MIDI patch bay manager for ALSA");
	g_option_context_add_main_entries(context, options, NULL);
	if (!g_option_context_parse(context, &argc, &argv, &error)) {
		abend(-1, "Failed to parse options: %s\n", error->message);
	}
	g_option_context_free(context);

	if (display_version) {
		print_version();
		exit(0);
	}

	// ensure a FILE was specified
	if (filenames == NULL || filenames[0] == NULL || filenames[1] != NULL) {
		abend(-1, "Exactly one FILE argument is required.\n");
	}

	// initialize
	initialize_alsa();
	Scripting L = initialize_scripting(filenames[0]);
	if (L == NULL) {
		abend(-1, "Failed to initialize scripting.\n");
	}

	load_script(L);
	reconfigure(L);

	if (listen) {
		listen_for_changes(L);
	}

	return 0;
}
