/*
 * patchbay.c - ALSA MIDI-related functions
 *
 * Copyright (C) 2014 Henry So, Jr.
 *
 * This file is part of Braze.
 *
 * Braze is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Braze is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Braze.  If not, see <http://www.gnu.org/licenses/>.
 */

#define _GNU_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <signal.h>
#include <poll.h>
#include <alsa/asoundlib.h>
#include "braze.h"

#define MY_CLIENT_NAME "Braze"

#define READ_CAPS (SND_SEQ_PORT_CAP_READ|SND_SEQ_PORT_CAP_SUBS_READ)
#define READ_MASK (SND_SEQ_PORT_CAP_READ|SND_SEQ_PORT_CAP_SUBS_READ|SND_SEQ_PORT_CAP_NO_EXPORT)
#define WRITE_CAPS (SND_SEQ_PORT_CAP_WRITE|SND_SEQ_PORT_CAP_SUBS_WRITE)
#define WRITE_MASK (SND_SEQ_PORT_CAP_WRITE|SND_SEQ_PORT_CAP_SUBS_WRITE|SND_SEQ_PORT_CAP_NO_EXPORT)

#define MODE_READ 1
#define MODE_WRITE 2

// globalize these to prevent having to allocate them all the time
static snd_seq_t *seq;
static snd_seq_client_info_t *cinfo;
static snd_seq_port_info_t *pinfo, *pinfo_sink;
static snd_seq_query_subscribe_t *query;
static snd_seq_port_subscribe_t *subs;

static int my_client;

extern gboolean listen;
extern gboolean show_system;
extern gboolean show_patches;
extern gboolean show_connections;

#ifndef NDEBUG
extern gboolean debug_memory;
#endif

// handle SIGHUP as a signal to reload
static volatile sig_atomic_t need_reload = FALSE;
static void handle_hup(int ignored) {
	need_reload = TRUE;
}

// initializes the signal handler
static void
setup_script_reload() {
	struct sigaction sa;
	sa.sa_handler = handle_hup;
	sigemptyset(&sa.sa_mask);
	sa.sa_flags = SA_RESTART;
	if (sigaction(SIGHUP, &sa, NULL) == -1) {
		// this is just reduced functionality, not fatal
		fprintf(stderr, "Warning: Failed to set up the signal handler. Cannot "
				"catch SIGHUP to reload");
	} 
}

// computes hash for snd_seq_addr_t
static guint
addr_hash(const snd_seq_addr_t *addr) {
	int hash = 5381;
	hash = (hash * 33) + addr->client;
	hash = (hash * 33) + addr->port;
	return hash;
}

// compares snd_seq_addr_t for equality
static gboolean
addr_equals(const snd_seq_addr_t *a, const snd_seq_addr_t *b) {
	return a->client == b->client && a->port == b->port;
}

// computes hash for port_locator_t
static guint
loc_hash(const port_locator_t *loc) {
	return (g_str_hash(loc->client_pattern) * 23) + g_str_hash(loc->port_pattern);
}

// compares port_locator_t for equality
static gboolean
loc_equals(const port_locator_t *a, const port_locator_t *b) {
	return g_str_equal(a->client_pattern, b->client_pattern) &&
		g_str_equal(a->port_pattern, b->port_pattern);
}

// frees memory used by port_locator_t
static void
loc_destroy(port_locator_t *loc) {
	if (loc->addresses != NULL) g_array_unref(loc->addresses);
	g_regex_unref(loc->port_re);
	g_regex_unref(loc->client_re);
	g_free(loc->port_pattern);
	g_free(loc->client_pattern);
	g_free(loc);
}

// computes hash for patch_t
static guint
patch_hash(const patch_t *patch) {
	return (loc_hash(patch->source) * 17) + loc_hash(patch->sink);
}

// compares patch_t for equality
static gboolean
patch_equals(const patch_t *a, const patch_t *b) {
	return loc_equals(a->source, b->source) && loc_equals(a->sink, b->sink);
}

// describes a port
typedef struct {
	guint mode;
	char name[6]; // 6 because the name will ultimately be [...]:[...]\0
	// the malloc'd structure will exceed sizeof(port_t) to hold the name
} port_t;

// describes a desired connection by its addresses
typedef struct {
	snd_seq_addr_t *source, *sink;
	patch_t *patch;
} connection_t;

// computes hash for connection_t
static guint
conn_hash(const connection_t *conn) {
	return (addr_hash(conn->source) * 17) + addr_hash(conn->sink);
}

// compares connection_t for equality
static gboolean
conn_equals(const connection_t *a, const connection_t *b) {
	return addr_equals(a->source, b->source) && addr_equals(a->sink, b->sink);
}

// matches using the regular expression, ignoring leading/trailing whitespace
static gboolean
re_match(const GRegex *re, const char *str) {
	while (isspace(*str)) ++str; // "trim" left
	GMatchInfo *match = NULL;
	gint start, end;
	if (g_regex_match(re, str, 0, &match)) {
		if (g_match_info_fetch_pos(match, 0, &start, &end) && start == 0) {
			g_match_info_unref(match);
			// check for only spaces after the end
			str += end;
			while (isspace(*str)) ++str;
			return *str == '\0';
		}
	}
	if (match != NULL) {
		g_match_info_unref(match);
	}
	return FALSE;
} 

gboolean
client_exists(const GRegex *client_re) {
	gboolean found = FALSE;
	int client;
	// look for a matching client name
	snd_seq_client_info_set_client(cinfo, -1);
	while (!found && snd_seq_query_next_client(seq, cinfo) >= 0) {
		client = snd_seq_client_info_get_client(cinfo);
		found = SND_SEQ_CLIENT_SYSTEM != client && my_client != client &&
			re_match(client_re, snd_seq_client_info_get_name(cinfo));
	}
	return found;
}

gboolean
port_exists(const GRegex *client_re, const GRegex *port_re, gboolean input) {
	guint caps = input? WRITE_CAPS : READ_CAPS;
	guint mask = caps | SND_SEQ_PORT_CAP_NO_EXPORT;

	gboolean found = FALSE;
	const char *client_name;
	// this is the standard ALSA sequencer loop to visit all ports of all client
	snd_seq_client_info_set_client(cinfo, -1);
	while (!found && snd_seq_query_next_client(seq, cinfo) >= 0) {
		int client = snd_seq_client_info_get_client(cinfo);
		// excluding myself...
		if (SND_SEQ_CLIENT_SYSTEM != client && my_client != client) {
			client_name = snd_seq_client_info_get_name(cinfo);
			snd_seq_port_info_set_client(pinfo, client);
			snd_seq_port_info_set_port(pinfo, -1);
			while (!found && snd_seq_query_next_port(seq, pinfo) >= 0) {
				found = (snd_seq_port_info_get_capability(pinfo) & mask) == caps &&
						re_match(client_re, client_name) &&
						re_match(port_re, snd_seq_port_info_get_name(pinfo));
			}
		}
	}
	return found;
}

// clones the address; will kill app if the allocation fails
static snd_seq_addr_t *
clone_addr(const snd_seq_addr_t *addr) {
	snd_seq_addr_t *cloned = g_new(snd_seq_addr_t, 1);
	cloned->client = addr->client;
	cloned->port = addr->port;
	return cloned;
}

// right-trims the string in-place and returns a pointer to the end of the string
static char *
trim_right_in_place(char *s) {
	char *end = s + strlen(s);
	while (--end >= s && isspace(*end)) ;
	*(++end) = '\0';
	return end;
}

// hash table visitor function that creates the addresses GArray
static void
create_address_array(gpointer ignore, port_locator_t *loc, gpointer ignore2) {
	loc->addresses = g_array_new(TRUE, FALSE, sizeof(snd_seq_addr_t));
	// g_array_new will never return NULL (program will die if alloc fails)
}

// gather port addresses corresponding to the locators and general port information
static void
gather_addresses_and_ports(GHashTable *locators, GHashTable *ports) {
	g_hash_table_foreach(locators, (GHFunc)create_address_array, NULL);

	int client;
	const char *client_name, *port_name;
	GHashTableIter iter;
	port_locator_t *loc;
	snd_seq_addr_t *addr;
	port_t *port;
	unsigned int caps;
	// this is the standard ALSA sequencer loop to visit all ports of all client
	snd_seq_client_info_set_client(cinfo, -1);
	while (snd_seq_query_next_client(seq, cinfo) >= 0) {
		client = snd_seq_client_info_get_client(cinfo);
		// excluding myself...
		if (SND_SEQ_CLIENT_SYSTEM != client && my_client != client) {
			snd_seq_port_info_set_client(pinfo, client);
			snd_seq_port_info_set_port(pinfo, -1);
			while (snd_seq_query_next_port(seq, pinfo) >= 0) {
				// only if the port is exported...
				if (!(snd_seq_port_info_get_capability(pinfo) & SND_SEQ_PORT_CAP_NO_EXPORT)) {
					client_name = snd_seq_client_info_get_name(cinfo);
					port_name = snd_seq_port_info_get_name(pinfo);
					addr = clone_addr(snd_seq_port_info_get_addr(pinfo));

					// iterate through all locators and add the address if it matches
					g_hash_table_iter_init(&iter, locators);
					while (g_hash_table_iter_next(&iter, (gpointer *)&loc, NULL)) {
						if (re_match(loc->client_re, client_name) &&
								re_match(loc->port_re, port_name)) {
							g_array_append_val(loc->addresses, *addr);
						}
					}

					// get general port information for later use
					caps = snd_seq_port_info_get_capability(pinfo);

					// advance pointers here to avoid over-allocating for leading spaces
					while (isspace(*client_name)) ++client_name; // "trim" left
					while (isspace(*port_name)) ++port_name; // "trim" left
					// this will over-allocate for trailing spaces
					port = g_malloc(sizeof(port_t) + strlen(client_name) + strlen(port_name));
					port->mode = 0;
					if ((caps & READ_CAPS) == READ_CAPS) {
						port->mode |= MODE_READ;
					}
					if ((caps & WRITE_CAPS) == WRITE_CAPS) {
						port->mode |= MODE_WRITE;
					}
					// advance name to reduce need to compute strlen all the time
					char *name = port->name;
					strcpy(name++, "[");
					strcpy(name, client_name);
					strcpy(name = trim_right_in_place(name), "]:[");
					strcpy(name += 3, port_name);
					strcpy(trim_right_in_place(name), "]");

					g_hash_table_insert(ports, addr, port);
				}
			}
		}
	}
}

// compute the connections based on the desired patches coming from the Lua script
static void
compute_connections(GHashTable *connections, GHashTable *patches, GHashTable *ports) {
	GHashTableIter iter;
	patch_t *patch;
	GArray *sources, *sinks;
	int i, j;
	connection_t key, *conn;
	snd_seq_addr_t *source, *sink;
	port_t *source_port, *sink_port;
	// iterate through all patches
	g_hash_table_iter_init(&iter, patches);
	while (g_hash_table_iter_next(&iter, (gpointer *)&patch, NULL)) {
		// get the addresses (filled in by reference when the addresses were gathered)
		sources = patch->source->addresses;
		sinks = patch->sink->addresses;

		// loop through all source addresses...
		for (i = 0; (key.source = &g_array_index(sources, snd_seq_addr_t, i))->client; i++) {
			source_port = g_hash_table_lookup(ports, key.source);
			// ...that are output ports
			if (source_port->mode & MODE_READ) {
				// loop through all sink addresses...
				for (j = 0; (key.sink = &g_array_index(sinks, snd_seq_addr_t, j))->client; j++) {
					sink_port = g_hash_table_lookup(ports, key.sink);
					// ... that are input ports
					if (sink_port->mode & MODE_WRITE) {
						// create the connection if it hasn't already been created
						if (!g_hash_table_contains(connections, &key)) {
							conn = g_new(connection_t, 1);
							conn->source = key.source;
							conn->sink = key.sink;
							conn->patch = patch;
							g_hash_table_add(connections, conn);

							if (show_patches) {
								if (conn->patch->time) {
									printf("Patching %s to %s %s:%d%s\n",
											source_port->name, sink_port->name,
											conn->patch->real? "real" : "tick",
											conn->patch->queue,
											conn->patch->exclusive? " excl" : "");
								}
								else {
									printf("Patching %s to %s%s\n",
											source_port->name, sink_port->name,
											conn->patch->exclusive? " excl" : "");
								}
							}
						}
					}
				}
			}
		}
	}
}

// removes any connections not in the given hash table
static void
remove_connections(GHashTable *connections, GHashTable *ports) {
	int client, queue;
	connection_t conn;
	const char *source_client, *source_port;
	// this is the standard ALSA sequencer loop to visit all ports of all client
	snd_seq_client_info_set_client(cinfo, -1);
	while (snd_seq_query_next_client(seq, cinfo) >= 0) {
		client = snd_seq_client_info_get_client(cinfo);
		// excluding myself...
		if (SND_SEQ_CLIENT_SYSTEM != client && my_client != client) {
			source_client = snd_seq_client_info_get_name(cinfo);
			snd_seq_port_info_set_client(pinfo, client);
			snd_seq_port_info_set_port(pinfo, -1);
			while (snd_seq_query_next_port(seq, pinfo) >= 0) {
				// only for output ports
				if ((snd_seq_port_info_get_capability(pinfo) & READ_CAPS) == READ_CAPS) {
					source_port = snd_seq_port_info_get_name(pinfo);
					// for all subscriptions from this output port
					snd_seq_query_subscribe_set_root(query, snd_seq_port_info_get_addr(pinfo));
					snd_seq_query_subscribe_set_type(query, SND_SEQ_QUERY_SUBS_READ);
					snd_seq_query_subscribe_set_index(query, 0);
					while (snd_seq_query_port_subscribers(seq, query) >= 0) {
						conn.source = (snd_seq_addr_t *)snd_seq_query_subscribe_get_root(query);
						conn.sink = (snd_seq_addr_t *)snd_seq_query_subscribe_get_addr(query);

						// skip if this is supposed to be connected or if unsubscribing failed
						if (g_hash_table_contains(connections, &conn) ||
								snd_seq_get_any_port_info(seq, conn.sink->client, conn.sink->port, pinfo_sink) < 0 ||
								(snd_seq_port_info_get_capability(pinfo_sink) & WRITE_CAPS) != WRITE_CAPS) {
							// remove the connection because it's already connected
							g_hash_table_remove(connections, &conn);
							snd_seq_query_subscribe_set_index(query, snd_seq_query_subscribe_get_index(query) + 1);
							continue;
						}
						// otherwise, attempt to unsubscribe
						queue = snd_seq_query_subscribe_get_queue(query);
						snd_seq_port_subscribe_set_queue(subs, queue);
						snd_seq_port_subscribe_set_sender(subs, conn.source);
						snd_seq_port_subscribe_set_dest(subs, conn.sink);
						if (snd_seq_unsubscribe_port(seq, subs) < 0) {
							// failed to unsubscribe
							snd_seq_query_subscribe_set_index(query, snd_seq_query_subscribe_get_index(query) + 1);
						}
						else {
							// unsubscribe successful
							if (show_connections) {
								if (snd_seq_query_subscribe_get_time_update(query)) {
									printf("Removed %s:%s -> %s %s:%d%s\n", source_client, source_port,
											((port_t *)g_hash_table_lookup(ports, conn.sink))->name,
											snd_seq_query_subscribe_get_time_real(query)? "real" : "tick", queue,
											snd_seq_query_subscribe_get_exclusive(query)? " excl" : "");
								}
								else {
									printf("Removed %s:%s -> %s%s\n", source_client, source_port,
											((port_t *)g_hash_table_lookup(ports, conn.sink))->name,
											snd_seq_query_subscribe_get_exclusive(query)? " excl" : "");
								}
							}
						}
					}
				}
			}
		}
	}
}

// hash table visitor function that adds a connection
static void
add_connection(connection_t *conn, gpointer ignored, GHashTable *ports) {
	snd_seq_port_subscribe_set_sender(subs, conn->source);
	snd_seq_port_subscribe_set_dest(subs, conn->sink);
	snd_seq_port_subscribe_set_queue(subs, conn->patch->queue);
	snd_seq_port_subscribe_set_exclusive(subs, conn->patch->exclusive);
	snd_seq_port_subscribe_set_time_update(subs, conn->patch->time);
	snd_seq_port_subscribe_set_time_real(subs, conn->patch->real);

	if (snd_seq_subscribe_port(seq, subs) >= 0) {
		// subscribe successful
		if (show_connections) {
			if (conn->patch->time) {
				printf("Added %s -> %s %s:%d%s\n",
						((port_t *)g_hash_table_lookup(ports, conn->source))->name,
						((port_t *)g_hash_table_lookup(ports, conn->sink))->name,
						conn->patch->real? "real" : "tick", conn->patch->queue,
						conn->patch->exclusive? " excl" : "");
			}
			else {
				printf("Added %s -> %s%s\n",
						((port_t *)g_hash_table_lookup(ports, conn->source))->name,
						((port_t *)g_hash_table_lookup(ports, conn->sink))->name,
						conn->patch->exclusive? " excl" : "");
			}
		}
	}
}

// called in response to client/port events to reconfigute the sequencer patches
void
reconfigure(Scripting L) {
	if (show_system) {
		printf("-- Reconfiguring --\n");
	}

	GHashTable *locators = g_hash_table_new_full((GHashFunc)loc_hash,
			(GEqualFunc)loc_equals, (GDestroyNotify)loc_destroy, NULL);
	GHashTable *patches = g_hash_table_new_full((GHashFunc)patch_hash,
			(GEqualFunc)patch_equals, g_free, NULL);

	setup_for_reconfigure(L, locators, patches);

	// call the Lua script
	if (call_reconfigure_script(L, locators, patches)) {
		GHashTable *ports = g_hash_table_new_full((GHashFunc)addr_hash,
				(GEqualFunc)addr_equals, g_free, g_free);
		gather_addresses_and_ports(locators, ports);

		GHashTable *connections = g_hash_table_new_full((GHashFunc)conn_hash,
				(GEqualFunc)conn_equals, g_free, NULL);
		compute_connections(connections, patches, ports);

		remove_connections(connections, ports);

		g_hash_table_foreach(connections, (GHFunc)add_connection, ports);

		g_hash_table_unref(connections);
		g_hash_table_unref(ports);
	}

	g_hash_table_unref(patches);
	g_hash_table_unref(locators);

	// remove access to the patch function defined above
	cleanup_after_reconfigure(L);

	if (show_system) {
		printf("\n");
	}
#ifndef NDEBUG
	if (debug_memory) {
		g_mem_profile();
	}
#endif
}

// shuts down ALSA connectivity
void
destroy_alsa() {
	snd_seq_close(seq);

	g_free(subs);
	g_free(query);
	g_free(pinfo_sink);
	g_free(pinfo);
	g_free(cinfo);

	seq = NULL;

	subs = NULL;
	query = NULL;
	pinfo_sink = NULL;
	pinfo = NULL;
	cinfo = NULL;
}

// initializes ALSA connectivity
void
initialize_alsa() {
	// allocate opaque ALSA structures from the stack for reuse
	cinfo = g_malloc(snd_seq_client_info_sizeof());
	pinfo = g_malloc(snd_seq_port_info_sizeof());
	pinfo_sink = g_malloc(snd_seq_port_info_sizeof());
	query = g_malloc(snd_seq_query_subscribe_sizeof());
	subs = g_malloc(snd_seq_port_subscribe_sizeof());

	if (snd_seq_open(&seq, "default", SND_SEQ_OPEN_INPUT, SND_SEQ_NONBLOCK) < 0) {
		abend(-1, "Failed to open sequencer.\n");
	}
	atexit(destroy_alsa);

	snd_seq_set_client_name(seq, MY_CLIENT_NAME);

	snd_seq_get_client_info(seq, cinfo);
	my_client = snd_seq_client_info_get_client(cinfo);

	// check if a different instance of Braze is registered as a client
	snd_seq_client_info_set_client(cinfo, -1);
	while (snd_seq_query_next_client(seq, cinfo) >= 0) {
		if (my_client != snd_seq_client_info_get_client(cinfo) &&
				strcmp(MY_CLIENT_NAME, snd_seq_client_info_get_name(cinfo)) == 0) {
			abend(-1, "Braze is currently running.\n");
		}
	}

	if (listen) {
		int myport = snd_seq_create_simple_port(seq, "Listener",
				SND_SEQ_PORT_CAP_WRITE|SND_SEQ_PORT_CAP_SUBS_WRITE|SND_SEQ_PORT_CAP_NO_EXPORT,
				SND_SEQ_PORT_TYPE_APPLICATION);

		if (myport < 0) {
			abend(-1, "Failed to create port.\n");
		}

		snd_seq_connect_from(seq, myport, SND_SEQ_CLIENT_SYSTEM, SND_SEQ_PORT_SYSTEM_ANNOUNCE);
	}
}

// endless loop to listen for changes to ALSA sequencer devices
void
listen_for_changes(Scripting L) {
	int size = snd_seq_poll_descriptors_count(seq, POLLIN);
	struct pollfd *pfd = (struct pollfd *)alloca(size * sizeof(struct pollfd));
	snd_seq_poll_descriptors(seq, pfd, size, POLLIN);
	sigset_t mask, omask;
	sigemptyset(&mask);
	sigaddset(&mask, SIGHUP);
	const struct timespec settle_timeout = { 0, 100000000 }; // 100ms

	int ready;
	gboolean is_reconfigure_event;
	snd_seq_event_t *event;

	// to avoid a race condition, we want to block SIGHUP except when polling
	sigprocmask(SIG_BLOCK, &mask, &omask);
	setup_script_reload(); // only start catching SIGHUP from this point
	for (;;) {
		is_reconfigure_event = FALSE;
		ready = ppoll(pfd, size, NULL, &omask);

		// this loop waits for the devices to settle
		do {
			if (ready > 0) {
				do {
					snd_seq_event_input(seq, &event);
					switch (event->type) {
					case SND_SEQ_EVENT_CLIENT_START:
					case SND_SEQ_EVENT_CLIENT_EXIT:
					case SND_SEQ_EVENT_CLIENT_CHANGE:
					case SND_SEQ_EVENT_PORT_START:
					case SND_SEQ_EVENT_PORT_EXIT:
					case SND_SEQ_EVENT_PORT_CHANGE:
						is_reconfigure_event = TRUE;
						break;
					}
					snd_seq_free_event(event);
				} while (snd_seq_event_input_pending(seq, 0) > 0);
			}
			ready = ppoll(pfd, size, &settle_timeout, &omask);
		} while (ready > 0);

		if (need_reload) {
			load_script(L);
			need_reload = FALSE;
			is_reconfigure_event = TRUE;
		}

		if (is_reconfigure_event) {
			reconfigure(L);
		}
	}
	// loop will never get here
	//sigprocmask(SIG_SETMASK, &omask, NULL);
}
