/*
 * scripting.c - Lua scripting functions
 *
 * Copyright (C) 2014 Henry So, Jr.
 *
 * This file is part of Braze.
 *
 * Braze is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Braze is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Braze.  If not, see <http://www.gnu.org/licenses/>.
 */

#define _GNU_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <lua.h>
#include <lauxlib.h>
#include <lualib.h>
#include "braze.h"

#define PATCH_CONFIG_MT "braze.patch"
#define RECONFIGURE_FN "reconfigure"
#define PATCH_FN "patch"

static gboolean is_first_run;
static char *script;

extern gboolean show_system;

// compiles a regular expression, sending any error to the Lua stack
static GRegex *
re_compile(lua_State *L, const char *pattern) {
	GError *error = NULL;
	GRegex *re = g_regex_new(pattern,
			G_REGEX_ANCHORED|G_REGEX_NO_AUTO_CAPTURE|G_REGEX_OPTIMIZE,
			G_REGEX_MATCH_NOTEMPTY, &error);

	if (re == NULL) {
		lua_pushfstring(L, "invalid regular expression /%s/: %s", pattern,
				error->message);
	}

	return re;
}

// Lua function: client(name_re) returns whether a client matches
static int
luafunc_client(lua_State *L) {
	GRegex *re = re_compile(L, luaL_checkstring(L, 1));
	if (re == NULL) {
		lua_error(L);
	}

	gboolean found = client_exists(re);

	g_regex_unref(re);
	lua_pushboolean(L, found);
	return 1;
}

// helper method for luafunc_inputport and luafunc_outputport
static int
match_port(lua_State *L, gboolean input) {
	GRegex *client_re = re_compile(L, luaL_checkstring(L, 1));
	if (client_re == NULL) {
		lua_error(L);
	}
	GRegex *port_re = re_compile(L, luaL_checkstring(L, 2));
	if (port_re == NULL) {
		g_regex_unref(client_re);
		lua_error(L);
	}

	gboolean found = port_exists(client_re, port_re, input);

	g_regex_unref(port_re);
	g_regex_unref(client_re);
	lua_pushboolean(L, found);
	return 1;
}

// Lua function: inputport(client_re, port_re) returns whether an input port matches
int
luafunc_inputport(lua_State *L) {
	return match_port(L, TRUE);
}

// Lua function: outputport(client_re, port_re) returns whether an output port matches
int
luafunc_outputport(lua_State *L) {
	return match_port(L, FALSE);
}

// clones the string; will return NULL if the allocation fails
static char *
clone_string(const char *s) {
	char *cloned = g_try_malloc(strlen(s) + 1);
	if (cloned != NULL) {
		strcpy(cloned, s);
	}
	return cloned;
}

// finds or creates a locator based on values in the Lua stack
static port_locator_t *
find_or_create_locator(lua_State *L, GHashTable *locators,
		const char *client_pattern, const char *port_pattern) {
	port_locator_t key, *loc;
	key.client_pattern = (char *)client_pattern;
	key.port_pattern = (char *)port_pattern;

	loc = g_hash_table_lookup(locators, &key);

	if (loc == NULL) {
		// not found, so create a new one
		loc = g_try_new(port_locator_t, 1);
		if (loc == NULL) {
			lua_pushstring(L, "Failed to allocate memory.");
			lua_error(L);
		}
		loc->client_pattern = clone_string(key.client_pattern);
		if (loc->client_pattern == NULL) {
			g_free(loc);
			lua_pushstring(L, "Failed to allocate memory.");
			lua_error(L);
		}
		loc->port_pattern = clone_string(key.port_pattern);
		if (loc->port_pattern == NULL) {
			g_free(loc->client_pattern);
			g_free(loc);
			lua_pushstring(L, "Failed to allocate memory.");
			lua_error(L);
		}
		loc->client_re = re_compile(L, loc->client_pattern);
		if (loc->client_re == NULL) {
			g_free(loc->port_pattern);
			g_free(loc->client_pattern);
			g_free(loc);
			// error message pushed by re_compile
			lua_error(L);
		}
		loc->port_re = re_compile(L, loc->port_pattern);
		if (loc->port_re == NULL) {
			g_regex_unref(loc->client_re);
			g_free(loc->port_pattern);
			g_free(loc->client_pattern);
			g_free(loc);
			// error message pushed by re_compile
			lua_error(L);
		}
		loc->addresses = NULL; // addresses will be filled in later

		g_hash_table_add(locators, loc);
	}

	return loc;
}

// userdata which contains intermediate data for creating a patch
typedef struct {
	GHashTable *locators, *patches;
	port_locator_t *source;
	int queue, time, real, exclusive;
} patch_config_t;

// Lua function: patch(...).to(client_re, port_re) 
// patches the source defined by patch(...) to the sink defined by the client_re and port_re
static int
luafunc_patch_to(lua_State *L) {
	// argument 1 is self
	patch_config_t *config = (patch_config_t *)luaL_checkudata(L, 1, PATCH_CONFIG_MT);
	GHashTable *patches = config->patches;

	patch_t key, *patch;
	key.source = config->source;
	key.sink = find_or_create_locator(L, config->locators,
			// argument 2 is the client regular expression pattern
			luaL_checkstring(L, 2),
			// argument 3 is the port regular expression pattern
			luaL_checkstring(L, 3));
	if (!g_hash_table_contains(patches, &key)) {
		patch = g_new(patch_t, 1);
		patch->source = key.source;
		patch->sink = key.sink;
		patch->queue = config->queue;
		patch->time = config->time;
		patch->real = config->real;
		patch->exclusive = config->exclusive;
		g_hash_table_add(patches, patch);
	}
	return 0;
}

// Lua function: patch(...).time(type[, queue]) 
static int
luafunc_patch_time(lua_State *L) {
	patch_config_t *config = (patch_config_t *)luaL_checkudata(L, 1, PATCH_CONFIG_MT);

	const char *type = NULL;
	if (!lua_isnil(L, 2)) {
		type = luaL_checkstring(L, 2);
	}

	if (type == NULL || strlen(type) == 0) {
		config->queue = 0;
		config->time = 0;
		config->real = 0;
	}
	else if (strcasecmp(type, "tick") == 0) {
		config->queue = luaL_checkint(L, 3);
		config->time = 1;
		config->real = 0;
	}
	else if (strcasecmp(type, "real") == 0) {
		config->queue = luaL_checkint(L, 3);
		config->time = 1;
		config->real = 1;
	}
	else {
		lua_pushfstring(L, "Invalid queue type: %s", type);
		lua_error(L);
	}

	lua_pushvalue(L, 1); // get the userdata back onto the stack
	return 1;
}

// Lua function: patch(...).exclusive([boolean]) 
// configures the patch to an exclusive connection
static int
luafunc_patch_exclusive(lua_State *L) {
	patch_config_t *config = (patch_config_t *)luaL_checkudata(L, 1, PATCH_CONFIG_MT);
	if (lua_isnone(L, 2)) {
		config->exclusive = 1;
	}
	else {
		luaL_checktype(L, 2, LUA_TBOOLEAN);
		config->exclusive = lua_toboolean(L, 2)? 1 : 0;
	}

	lua_pushvalue(L, 1); // get the userdata back onto the stack
	return 1;
}

// Lua function: patch(client_re, port_re)
// defines the source using client_re and port_re and readies it for a call to to(...)
static int
luafunc_patch(lua_State *L) {
	patch_config_t *config = (patch_config_t *)lua_newuserdata(L, sizeof(patch_config_t));
	luaL_getmetatable(L, PATCH_CONFIG_MT);
	lua_setmetatable(L, -2);

	// upvalue 1 is the set of port locators
	config->locators = (GHashTable *)lua_touserdata(L, lua_upvalueindex(1));
	// upvalue 2 is the set of patches to define
	config->patches = (GHashTable *)lua_touserdata(L, lua_upvalueindex(2));
	config->source = find_or_create_locator(L, config->locators,
			// argument 1 is the client regular expression pattern
			luaL_checkstring(L, 1),
			// argument 2 is the port regular expression pattern
			luaL_checkstring(L, 2));
	config->queue = 0;
	config->time = 0;
	config->real = 0;
	config->exclusive = 0;
	
	return 1; // config is still on the stack
}

// sets up the Lua environment for a reconfigure cycle
void
setup_for_reconfigure(Scripting L, GHashTable *locators, GHashTable *patches) {
	lua_pushlightuserdata(L, locators); // as outgoing upvalue 1
	lua_pushlightuserdata(L, patches); // as outgoing upvalue 2
	lua_pushcclosure(L, luafunc_patch, 2);
	lua_setglobal(L, PATCH_FN); // define patch function
}

// calls the reconfigure script loaded in the current Lua context
gboolean
call_reconfigure_script(Scripting L, GHashTable *locators, GHashTable *patches) {
	lua_getglobal(L, RECONFIGURE_FN);
	int rc = lua_pcall(L, 0, 1, 0);

	if (rc) {
		fprintf(stderr, "Failed to reconfigure: %s\n", lua_tostring(L, -1));
		lua_pop(L, 1); // pop error message off
		return FALSE;
	}

	if (is_first_run) {
		is_first_run = FALSE;

		// handle case where script needs to initialize; in this case,
		// the script should return the real reconfiguration function
		if (lua_isfunction(L, -1)) {
			lua_setglobal(L, RECONFIGURE_FN);
			g_hash_table_remove_all(locators);
			g_hash_table_remove_all(patches);

			if (show_system) {
				printf("-- Initialized %s --\n", script);
			}

			return call_reconfigure_script(L, locators, patches);
		}
	}

	lua_pop(L, 1);

	return TRUE;
}

// cleans up after a reconfigure cycle
void
cleanup_after_reconfigure(Scripting L) {
	// remove access to the patch function defined above
	lua_pushnil(L);
	lua_setglobal(L, PATCH_FN);
}

// binds a global C function into Lua
static void
lglobal_cfunction(lua_State *L, char *name, lua_CFunction fn) {
	lua_pushcfunction(L, fn);
	lua_setglobal(L, name);
}

// shuts down Lua
static lua_State *lua_State_to_close;
void
destroy_lua() {
	lua_close(lua_State_to_close);
	lua_State_to_close = NULL;
}

static const struct luaL_reg patch_config_reg[] = {
	{ "time", luafunc_patch_time },
	{ "exclusive", luafunc_patch_exclusive },
	{ "to", luafunc_patch_to },
	{ NULL, NULL },
};

// initializes Lua
lua_State *
initialize_scripting(char *filename) {
	script = filename;

	lua_State *L = lua_State_to_close = luaL_newstate();
	if (L != NULL) {
		atexit(destroy_lua);

		// set up the metatable for patch_config_t
		luaL_newmetatable(L, PATCH_CONFIG_MT);
		lua_pushstring(L, "__index");
		lua_pushvalue(L, -2); // the metatable
		lua_settable(L, -3); // metatable.__index = metatable
		luaL_openlib(L, NULL, patch_config_reg, 0);

		luaL_openlibs(L);

		// bind functions
		lglobal_cfunction(L, "client", luafunc_client);
		lglobal_cfunction(L, "inputport", luafunc_inputport);
		lglobal_cfunction(L, "outputport", luafunc_outputport);
	}

	return L;
}

// loads and binds the Lua script to the Lua function reconfigure
void
load_script(lua_State *L) {
	if (show_system) {
		printf("-- Loading script from %s --\n", script);
	}

	if (luaL_loadfile(L, script)) {
		fprintf(stderr, "Failed to load: %s\n", lua_tostring(L, -1));
		lua_pop(L, 1); // pop error message off
		exit(-1);
	}
	lua_setglobal(L, RECONFIGURE_FN);
	is_first_run = TRUE;
}

